{$S-,R-,V-,I-,B-,F+}

{Conditional defines that may affect this unit}
{$I BTDEFINE.INC}

{$IFNDEF Ver40}
{$I OPLUS.INC}                             {Allow overlays}
{$I AMINUS.INC} {Avoid byte mismatches for keyboard installation}
{$ENDIF}

{*********************************************************}
{*                   BROWSER.PAS 5.00                    *}
{*    Copyright (c) Dipl.Math. Ralf Nagel 1986/87/88.    *}
{*                 All rights reserved.                  *}
{*          Modified and used under license by           *}
{*                 TurboPower Software.                  *}
{*                                                       *}
{*********************************************************}
{*********************************************************}
{*                   BROWSERM.PAS 1.00                   *}
{*                                                       *}
{*             Modified to add Mouse support             *}
{*     Donated to the Public Domain By Randall L. Smith  *}
{*                                                       *}
{*                                                       *}
{*********************************************************}

unit BrowserM;

interface

uses
{.U-}             {PSA directive don't do other units!}
    Printer,                                { test only!! }
    TpString,                               { test only!! }
    TpCmd,
    {$IFDEF UseTpMouse}
    TpWindow,                           { use window unit }
    TpMouse,                             { and mouse unit }
    {$ENDIF}

    {$IFDEF UseTPCRT}
    TpCrt,
    {$ELSE}
    Crt,
{$ENDIF}
    Dos,
    Filer,
    VRec;
{.U+}                                   {PSA directive}
const
    MinRows        = 4; {Minimum number of usable lines on the screen}
    MaxRows        = 20; {Maximum number of usable lines on the screen}
    MaxCols        = 128; {Number of usable columns on screen, must be <= 255}

    {Number of retries for a read operation in case of a Lock-Error}
    RetriesOnLock  : integer = 50;



    {True lets browser go to disk less often, safe only for single user mode}
    NoNetMode      : boolean = false;

type
    RowRec =
        record
            IKS            : IsamKeyStr;
            Ref            : longint;
            Row            : string[MaxCols];
        end;
  {Basic type of the browser. The browser is used in three steps:
   1) <IKS> and <Ref> are allocated.
   2) <ProcBuildaRow> (procedure parameter) determines the row.
   3) <ProcDisplayaRow> (procedure parameter) puts <Row> on the screen}

    {----------------- programmer's hooks -------------}
var
    BrowseKeyPtr   : pointer; {pointer to routine to return next keystroke}
    BrowseHelpPtr  : pointer; {pointer to routine to display help}
    BrowseHelpIndex : word;            {current help topic}

    {.F-}
const
    BKnone      = 00; {Not a command}
    BKchar      = 01; {Regular character--not a command}
    BKenter     = 02; {Select}
    BKquit      = 03; {Escape}
    BKfirstRec  = 04; {Cursor to first record}
    BKlastRec   = 05; {Cursor to last record}
    BKleft      = 06; {Cursor left one column}
    BKright     = 07; {Cursor right one column}
    BKup        = 08; {Cursor up one row}
    BKdown      = 09; {Cursor down one row}
    BKpageUp    = 10; {Cursor up one page}
    BKpageDown  = 11; {Cursor down one page}
    BKplus      = 12; {Reread current record}
    BKhelp      = 13; {Invoke help routine}

    BKtask0     = 14; {user-defined task commands}
    BKtask1     = 15;
    BKtask2     = 16;
    BKtask3     = 17;
    BKtask4     = 18;
    BKtask5     = 19;
    BKtask6     = 20;
    BKtask7     = 21;
    BKtask8     = 22;
    BKtask9     = 23;
    BKuser0     = 24; {user-defined exit commands}
    BKuser1     = 25;
    BKuser2     = 26;
    BKuser3     = 27;
    BKuser4     = 28;
    BKuser5     = 29;
    BKuser6     = 30;
    BKuser7     = 31;
    BKuser8     = 32;
    BKuser9     = 33;
    BKProbe     = 34;
type
    BKtype = BKnone..BKProbe;
const
    {$IFDEF UseTPCRT}
    HelpForBrowse = TpCrt.HelpForXXXX2; {= 7}
    {$ELSE}
    HelpForBrowse = 7;
    {$ENDIF}

    {Keystroke to command mapping}
    BrowseKeyMax = 212;   {last available slot in BrowseKeySet}

    {ID string for installation programs}
    BrowseKeyID : string[17] = 'browser key array';

    {default key assignments}
    BrowseKeySet : array[0..BrowseKeyMax] of Byte = (
       {length keys         command type     key sequence}
        3,     $00, $00,    BKquit,          {^Break}
        3,     $00, $3B,    BKhelp,          {F1}
        3,     $00, $47,    BKfirstRec,      {Home}
        3,     $00, $48,    BKup,            {Up}
        3,     $00, $49,    BKpageUp,        {PgUp}
        3,     $00, $4B,    BKleft,          {Left}
        3,     $00, $4D,    BKright,         {Right}
        3,     $00, $4F,    BKlastRec,       {End}
        3,     $00, $50,    BKdown,          {Down}
        3,     $00, $51,    BKpageDown,      {PgDn}
        3,     $00, $76,    BKlastrec,       {^PgDn}
        3,     $00, $84,    BKfirstrec,      {^PgUp}
        2,     $03,         BKpageDown,      {^C}
        2,     $04,         BKright,         {^D}
        2,     $05,         BKup,            {^E}
        2,     $0D,         BKenter,         {^M, Enter}
        2,     $12,         BKpageUp,        {^R}
        2,     $13,         BKleft,          {^S}
        2,     $17,         BKup,            {^W}
        2,     $18,         BKdown,          {^X}
        2,     $1A,         BKdown,          {^Z}
        2,     $1B,         BKquit,          {Esc}
        2,     $2B,         BKplus,          {+}
        3,     $11, $03,    BKlastrec,       {^Q^C}
        3,     $11, $12,    BKfirstrec,      {^Q^R}
{$IFDEF UseTpMouse}
        3,     $00, $EF,    BKProbe,         {Click left}
        3,     $00, $EE,    BKQuit,          {Click right}
        3,     $00, $ED,    BKHelp,          {Click both}
{$ELSE}
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0, 0, 0,
{$ENDIF}
        {-----------pad to end of array----------}
                                0, 0,        {90}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {100}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {110}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {120}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {130}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {140}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {150}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {160}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {170}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {180}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,        {190}
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0);       {200}

    {$IFDEF UseTpMouse}
    {Mouse control for Browse}
    BrowseMouseScroll : boolean = false; {True if scrolling available}
    BrowseMousePage : boolean = false; {True to scroll by one page per click}
    BrowseMouseEnabled : boolean = false; {True if mouse is enabled}
    MouseUpMark    = #24;        {Characters in scroll bar}
    MouseDnMark    = #25;
    MouseUpCmd     : array[boolean] of BKtype = (BKup, BKpageUp);
    MouseDownCmd   : array[boolean] of BKtype = (BKdown, BKpageDown);

var
    W              : WindowPtr;             {browse window}
    NrOfRows       : integer;



    {$ENDIF}





{.F+}
function BrowseM(IFBPtr         : IsamFileBlockPtr;
                 VarRec         : boolean;
                 KeyNr          : integer;
                 LowKey         : IsamKeyStr;
                 HighKey        : IsamKeyStr;

                 C1, R1, C2, R2 : integer;
                 Frame          : boolean;
                 Wattr, Fattr, Tattr : byte;
                 Title          : string;

                 var DatS;
                 var DatLen     : word;
                 var Ref        : longint;
                 var KeyStr     : IsamKeyStr;
                 var ExitKey    : BKtype;
                 ProcSpecialTask : pointer;
                 ProcBuildaRow  : pointer;
                 ProcDisplayaRow : pointer) : integer;
    {-A list is created with the corresponding keys from <LowKey> to <HighKey>
      inclusive and their data structures.

      The parameter <VarRec> must be true for variable record lengths
      (otherwise false).

      The display starts at the row <StartScreenRow> and entails <NrOfRows>
      rows. <NrOfRows> may not be less than 4.

      <DatS> is a buffer for the data structure.

      <DatLen> contains the length of the current variable length record only
      if <VarRec> is true.

      <ExitKey> returns the code for the command that exited the Browser. In
      addition this determines the first action the Browser should perform.
      BKnone (Byte 0) must be passed if no action is to be performed.

      The procedures whose addresses are passed through the pointer-parameters
      <ProcBuildaRow> and <ProcDisplayaRow> are used to create and display a
      row. These procedures must be supplied by the user of Browser and must
      be declared as in the example procedure definition. These procedures
      must be declared as "FAR." This is either done through the use of the
      compiler directive $F+, or through the exportation from another Unit.
      Handling it in any other way will produce a program crash.

      There are three possible return values:
        0: No error encountered.
           <ExitKey> = BKquit  : Esc was pressed.
           <ExitKey> <> BKquit : <DatS> contains the data structure,
                                 <Ref> the corresponding reference,
                                 <KeyStr> the chosen key.
        1: There were no keys available that were in the requested range.
        2: Hard Error by Isam-Access (Class 2 or higher).

      Browser has the following default actions for each command:
        BKquit     : exit Browser without making a selection
        BKenter    : exit Browser making a choice
        BKfirstRec : scroll to the first data structure
        BKlastRec  : scroll to the last data structure
        BKpageUp   : scroll back a page
        BKpageDown : scroll forward a page
        BKup       : scroll up a row
        BKdown     : scroll down a row

      If <ProcDisplayaRow> points to a procedure that makes use of the
      parameter <HorizOfs>, horizontal scrolling is available through the
      following commands:
        BKleft  : Move the display window towards the left
        BKright : Move the display window towards the right.}

function AddBrowseCommand(Cmd : BKtype; NumKeys : byte; Key1, Key2 : word) : boolean;
    {-Add a new command key assignment or change an existing one}

procedure BuildaRow(var RR         : RowRec;
                    KeyNr          : integer;
                    var DatS;
                    DatLen         : word);
    {-This routine is a model declaration for the procedure parameter
      <ProcBuildARow> referenced in function <Browse>. This procedure is
      called to create a row with maximum number of columns <MaxCols> for the
      data structure and/or the key. This can be greater than the maximum
      count of screen columns if there is horizontal scrolling available.
      Important for Network use: if there is a lock on the desired data
      structure, it cannot be loaded, and <RR.Ref> = -1. An example is
      described below.}

procedure DisplayaRow(var RR         : RowRec;
                      KeyNr,
                      RowNr,
                      StartRow       : integer;
                      HighLight      : boolean;
                      var HorizOfs   : integer);
    {-This routine is a model declaration for the procedure parameter
      <ProcDisplayARow> referenced in function <Browse>. This procedure is
      called to display the rows constructed by <ProcBuildaRow>. <RowNr> gives
      the relative position of the row. <HighLight> = True means that this
      corresponds to the actual row, which should be written in some special
      way here. <HorizOfs> may be used to enable a horizontal scrolling. The
      value of this parameter lies between 0 and a maximum value set by
      <ProcDisplayaRow>. <ProcDisplayaRow> can simply ignore the parameter if
      horizontal scrolling is not desired. An example is described below.}

procedure SpecialTask(IFBPtr         : IsamFileBlockPtr;
                      var DatS;
                      Ref            : longint;
                      IKS            : IsamKeyStr;
                      KeyNr          : integer;
                      var Command    : BKtype;
                      var exitcode   : integer;
                      DatLen         : word);
    {-This routine is a model declaration for the procedure parameter
      <ProcSpecialTask> referenced in function <Browse>. On entry to this
      procedure, <DatS>, <Ref> and <IKS> point to the present choice.

      <Command> is the browser command that led to this procedure being
      called. <Command> does have another special meaning: it may return
      another valid command and <Browse> will perform the specified action. If
      the returned value of <Command> is also a user task command, the special
      task is called again. If no such action is desired, the special task
      must return <BKnone> (=0). Be careful to avoid infinite loops, which
      will occur if you don't change the value in <Command>.

      <ExitCode> is used to let Browser know whether or not an error was
      encountered. No Error is represented by 0. Any different code leads to
      the termination of Browser with this value as the function return code.}



function BrowseReadKey : word;
    {-Return next keystroke as a word}

    {$IFDEF UseTpMouse}
procedure EnableBrowseMouse;
    {-Enable mouse control of browse lists}
procedure DisableBrowseMouse;
    {-Disable mouse control of browse lists}
    {$ENDIF}


    {=========================================================================}

implementation


{$IFDEF UseTpMouse}
    {$I BROWSERM.INC }
    {$ENDIF}






    procedure HelpRoutine(UnitCode : byte; IdPtr : pointer; HelpIndex : word);
        {-Call routine pointed to by BrowseHelpPtr}
    inline(
        $FF / $1E / >BrowseHelpPtr); {call dword ptr [>BrowseHelpPtr]}

    function GetKey : word;
        {-Call routine pointed to by BrowseKeyPtr}
    inline(
        $FF / $1E / >BrowseKeyPtr); {call dword ptr [>BrowseKeyPtr]}



    function BrowseReadKey : word;
        {-Return next keystroke as a word}
    var
        Ch             : char;
    begin
        Ch := ReadKey;
        if Ch <> #0 then
            BrowseReadKey := ord(Ch)
        else
            BrowseReadKey := ord(ReadKey) shl 8;
    end;




    function BrowseI(IFBPtr         : IsamFileBlockPtr;
                     VarRec         : boolean;
                     KeyNr          : integer;
                     LowKey,
                     HighKey        : IsamKeyStr;
                     StartScreenRow,
                     NrOfRows       : integer;
                     var DatS;
                     var DatLen     : word;
                     var Ref        : longint;
                     var KeyStr     : IsamKeyStr;
                     var ExitKey    : BKtype;
                     ProcSpecialTask : pointer;
                     ProcBuildaRow  : pointer;
                     ProcDisplayaRow : pointer;
                     NetUsed        : boolean) : integer;
    type
        RowType = 1..MaxRows;
    var
        CurRow         : RowType;   {Present Screen Row}
        Result,
        Moved,
        HorizOfs       : integer;
        Action         : BKtype;
        ChWord         : word;
        LowRef         : longint;
        HighRef        : longint;
        BrowScreen     : array[1..MaxRows] of RowRec; {Complete Screen}

        {$IFDEF UseTpMouse}

        SaveMX : Byte;                {Saved mouse state}
        SaveMY : Byte;
        SaveMXL : Byte;
        SaveMXH : Byte;
        SaveMYL : Byte;
        SaveMYH : Byte;

        SaveWaitFor : Boolean;        {Saved WaitForButtonRelease variable}
        SaveMouseOn : Boolean;        {Was mouse cursor on at entry}
        {$ENDIF}






        procedure CallBuildaRow(UserRoutine    : pointer;
                                var RR         : RowRec;
                                KeyNr          : integer;
                                var DatS;
                                DatLen         : word);

            procedure CallUserRoutine(var RR         : RowRec;
                                      KeyNr          : integer;
                                      var DatS;
                                      DatLen         : word);
                {-Call UserRoutine with an action code}
            inline($FF / $5E / <UserRoutine); {call far dword ptr [bp+<UserRoutine]}

        begin
            CallUserRoutine(RR, KeyNr, DatS, DatLen);
        end;


        procedure CallDisplayaRow(UserRoutine    : pointer;
                                  var RR         : RowRec;
                                  KeyNr,
                                  RowNr,
                                  StartRow       : integer;
                                  HighLight      : boolean;
                                  var HorizOfs   : integer);

            procedure CallUserRoutine(var RR         : RowRec;
                                      KeyNr,
                                      RowNr,
                                      StartRow       : integer;
                                      HighLight      : boolean;
                                      var HorizOfs   : integer);
                {-Call UserRoutine with an action code}
            inline($FF / $5E / <UserRoutine); {call far dword ptr [bp+<UserRoutine]}

        begin
            CallUserRoutine(RR, KeyNr, RowNr, StartRow, HighLight, HorizOfs);
        end;


        procedure CallSpecialTask(UserRoutine    : pointer;
                                  IFBPtr         : IsamFileBlockPtr;
                                  var DatS;
                                  Ref            : longint;
                                  IKS            : IsamKeyStr;
                                  KeyNr          : integer;
                                  var Command    : BKtype;
                                  var exitcode   : integer;
                                  DatLen         : word);

            procedure CallUserRoutine(IFBPtr         : IsamFileBlockPtr;
                                      var DatS;
                                      Ref            : longint;
                                      IKS            : IsamKeyStr;
                                      KeyNr          : integer;
                                      var Command    : BKtype;
                                      var exitcode   : integer;
                                      DatLen         : word);
                {-Call UserRoutine with an action code}
            inline($FF / $5E / <UserRoutine); {call far dword ptr [bp+<UserRoutine]}

        begin
            CallUserRoutine(IFBPtr, DatS, Ref, IKS, KeyNr, Command, exitcode, DatLen);
        end;







        procedure EvaluateMousePosition(W : WindowPtr; var Cmd : BKType;
                                        Var CurRow : RowType);
            {-Evaluate effect of a mouse selection}
        var
            mX             : byte; {Mouse absolute X position}
            mY             : byte; {Mouse absolute Y position}
            mCol           : word; {Logical column where mouse is located}
            mChoice        : word; {Possible choice using mouse}
            MaxXinItem     : integer; {Highest column to select item with mouse}
        begin                   {  EvaluateMousePosition  }
            if BrowseMouseEnabled then
            begin

                {Compute absolute mouse coordinates}
                mX := MouseXLo + MouseKeyWordX;
                mY := MouseYLo + MouseKeyWordY;


                with WindowP(W)^, Draw do
                    if BrowseMouseScroll and (mX = XH1) then
                    begin
                        {In the scroll bar region, scroll to new row}
                        if mY = YL1 then
                            {Treat like Up arrow or PgUp}
                            Cmd := MouseUpCmd[BrowseMousePage]
                        else if mY = YH1 then
                            {Treat like Down arrow or PgDn}
                            Cmd := MouseDownCmd[BrowseMousePage];
                    end
                    else
                        if (mY >= YL) and (mY <= YH) then
                        begin
                            {In the active pick region, convert to window-relative}
                            dec (mY, YL);
                            if mY = CurRow then
                            {Select the item on the second click}
                                Cmd := BKEnter
                            else
                            begin
                                {Move the highlight over the item}
                                CurRow := mY;
                                Cmd  := BkNone;
                            end;
                        end; { if (mY >= YL) and (mY <= YH) }

            end; { if (mX = XH1) }
        end;                    {  EvaluateMousePosition  }





        procedure S_Key(IFBPtr         : IsamFileBlockPtr;
                        Key            : integer;
                        var UserDatRef : longint;
                        var UserKey    : IsamKeyStr);
        var
            RT             : integer;
        begin
            RT := 0;
            repeat
                if NetUsed then
                    SearchNetKey(IFBPtr, Key, UserDatRef, UserKey)
                else
                    SearchKey(IFBPtr, Key, UserDatRef, UserKey);
                inc(RT);
            until (RT >= RetriesOnLock) or (IsamErrorClass <> 2);
        end;


        procedure N_Key(IFBPtr         : IsamFileBlockPtr;
                        Key            : integer;
                        var UserDatRef : longint;
                        var UserKey    : IsamKeyStr);

        var
            RT             : integer;

        begin
            RT := 0;
            repeat
                if NetUsed then
                    NextNetKey(IFBPtr, Key, UserDatRef, UserKey)
                else
                    NextKey(IFBPtr, Key, UserDatRef, UserKey);
                inc(RT);
            until (RT >= RetriesOnLock) or (IsamErrorClass <> 2);
        end;

        procedure P_Key(IFBPtr         : IsamFileBlockPtr;
                        Key            : integer;
                        var UserDatRef : longint;
                        var UserKey    : IsamKeyStr);
        var
            RT             : integer;
        begin
            RT := 0;
            repeat
                if NetUsed then
                    PrevNetKey(IFBPtr, Key, UserDatRef, UserKey)
                else
                    PrevKey(IFBPtr, Key, UserDatRef, UserKey);
                inc(RT);
            until (RT >= RetriesOnLock) or (IsamErrorClass <> 2);
        end;

        procedure F_Key(IFBPtr         : IsamFileBlockPtr;
                        Key            : integer;
                        var UserDatRef : longint;
                        var UserKey    : IsamKeyStr;
                        NFSD           : integer);
        var
            RT             : integer;
        begin
            RT := 0;
            repeat
                if NetUsed then
                    FindNetKeyAndRef(IFBPtr, Key, UserDatRef, UserKey, NFSD)
                else
                    FindKeyAndRef(IFBPtr, Key, UserDatRef, UserKey, NFSD);
                inc(RT);
            until (RT >= RetriesOnLock) or (IsamErrorClass <> 2);
        end;

        function LastRow : integer;
        var
            I              : integer;
        begin
            I := MaxRows;
            while (I > 0) and (BrowScreen[I].Ref = longint(0)) do
                dec(I);
            LastRow := I;
        end;

        procedure EmptyBrowScreen;
        var
            I              : integer;
        begin
            for I := 1 to MaxRows do
                with BrowScreen[I] do
                begin
                    IKS := '';
                    Ref := longint(0);
                    Row := '';
                end;
        end;

        function KeyInBounds(KeyStr : IsamKeyStr) : boolean;
        begin
            KeyInBounds := false;
            if copy(KeyStr, 1, length(LowKey)) < LowKey then
                exit;
            if copy(KeyStr, 1, length(HighKey)) > HighKey then
                exit;
            KeyInBounds := true;
        end;

        procedure PlaceAtEnd(RR : RowRec);
        var
            I              : integer;
        begin
            I := LastRow;
            if I < NrOfRows then
                BrowScreen[I] := RR
            else
            begin
                for I := 1 to NrOfRows - 1 do
                    BrowScreen[I] := BrowScreen[I + 1];
                BrowScreen[NrOfRows] := RR;
            end;
        end;

        procedure PlaceAtTop(RR : RowRec);
        var
            I              : integer;
        begin
            for I := NrOfRows downto 2 do
                BrowScreen[I] := BrowScreen[I - 1];
            BrowScreen[1] := RR;
        end;

        function CompleteaPage : boolean;
            {-Load data records and build display strings for current screen page}
        var
            I              : integer;
        begin
            CompleteaPage := false;
            for I := 1 to NrOfRows do
            begin
                with BrowScreen[I] do
                begin
                    if Ref <> longint(0) then 
                    begin
                        if Ref <> longint(- 1) then 
                        begin
                            if VarRec then
                                GetVariableRec(IFBPtr, Ref, DatS, DatLen, Normal)
                            else if NetUsed then
                                GetNetRec(IFBPtr, Ref, DatS)
                            else
                                GetRec(IFBPtr, Ref, DatS);
                            if IsamErrorClass > 2 then
                                exit;
                        end;
                        if IsamErrorClass = 2 then
                            Ref := longint(- 1);
                        CallBuildaRow(ProcBuildaRow, BrowScreen[I], KeyNr, DatS, DatLen);
                    end;
                end;
            end;
            CompleteaPage := true;
        end;

        procedure DisplayaPage;
        var
            I              : integer;
        begin
            for I := 1 to NrOfRows do
                with BrowScreen[I] do
                    CallDisplayaRow(ProcDisplayaRow, BrowScreen[I], KeyNr, I,
                                    StartScreenRow, (I = CurRow), HorizOfs);
        end;

        function BuildFirstPage : integer;
        var
            I              : integer;
            TRec           : RowRec;
        begin
            BuildFirstPage := 2;
            with TRec do
            begin
                IKS := LowKey;
                S_Key(IFBPtr, KeyNr, Ref, IKS);
                if IsamErrorClass > 1 then
                    exit;
            end;
            if (IsamErrorClass = 1) or (not KeyInBounds(TRec.IKS)) then
            begin
                BuildFirstPage := 1;
                exit;
            end;
            I := 1;
            while (I <= NrOfRows) and (KeyInBounds(TRec.IKS)) and (IsamErrorClass = 0) do
            begin
                BrowScreen[I] := TRec;
                with TRec do 
                begin
                    inc(I);
                    if I <= NrOfRows then
                    begin
                        N_Key(IFBPtr, KeyNr, Ref, IKS);
                        if IsamErrorClass > 1 then
                            exit;
                    end;
                end;
            end;
            BuildFirstPage := 0;
        end;

        function BuildLastPage : integer;
        var
            I              : integer;
            TRec           : RowRec;
        begin
            BuildLastPage := 2;
            with TRec do 
            begin
                IKS := HighKey;
                while length(IKS) < MaxKeyLen do
                    IKS := IKS + #255;
                Ref := $0FFFFFFF;
                F_Key(IFBPtr, KeyNr, Ref, IKS, - 1);
                if IsamErrorClass > 1 then
                    exit;
            end;
            if (IsamErrorClass = 1) or (not KeyInBounds(TRec.IKS)) then
            begin
                BuildLastPage := 1;
                exit;
            end;
            I := 1;
            while (I <= NrOfRows) and (KeyInBounds(TRec.IKS)) and (IsamErrorClass = 0) do
            begin
                PlaceAtTop(TRec);
                with TRec do
                begin
                    inc(I);
                    if I <= NrOfRows then
                    begin
                        P_Key(IFBPtr, KeyNr, Ref, IKS);
                        if IsamErrorClass > 1 then
                            exit;
                    end;
                end;
            end;
            BuildLastPage := 0;
        end;

        function BuildNextPage(Nr : integer; var Moved : integer) : integer;
        var
            I              : integer;
            TRec           : RowRec;
        begin
            BuildNextPage := 2;
            I := LastRow;
            if I = 0 then
            begin
                BuildNextPage := 1;
                exit;
            end;
            TRec := BrowScreen[I];
            with TRec do 
            begin
                F_Key(IFBPtr, KeyNr, Ref, IKS, 0);
                if IsamErrorClass > 1 then
                    exit;
                N_Key(IFBPtr, KeyNr, Ref, IKS);
                if IsamErrorClass > 1 then
                    exit;
            end;
            I := 1;
            while (I <= Nr) and (KeyInBounds(TRec.IKS)) and (IsamErrorClass = 0) do
            begin
                PlaceAtEnd(TRec);
                with TRec do 
                begin
                    inc(I);
                    if I <= Nr then
                    begin
                        N_Key(IFBPtr, KeyNr, Ref, IKS);
                        if IsamErrorClass > 1 then
                            exit;
                    end;
                end;
            end;
            BuildNextPage := 0;
            Moved := I - 1;
        end;

        function BuildPrevPage(Nr : integer; var Moved : integer) : integer;
        var
            I              : integer;
            TRec           : RowRec;
        begin
            BuildPrevPage := 2;
            if BrowScreen[1].Ref = longint(0) then
            begin
                BuildPrevPage := 1;
                exit;
            end;
            TRec := BrowScreen[1];
            with TRec do
            begin
                F_Key(IFBPtr, KeyNr, Ref, IKS, 0);
                if IsamErrorClass > 1 then
                    exit;
                if IsamErrorClass = 1 then
                    F_Key(IFBPtr, KeyNr, Ref, IKS, - 1)
                else
                    P_Key(IFBPtr, KeyNr, Ref, IKS);
                if IsamErrorClass > 1 then
                    exit;
            end;
            I := 1;
            while (I <= Nr) and (KeyInBounds(TRec.IKS)) and (IsamErrorClass = 0) do
            begin
                PlaceAtTop(TRec);
                with TRec do
                begin
                    inc(I);
                    if I <= Nr then
                    begin
                        P_Key(IFBPtr, KeyNr, Ref, IKS);
                        if IsamErrorClass > 1 then
                            exit;
                    end;
                end;
            end;
            Moved := I - 1;
            BuildPrevPage := 0;
        end;

        function BuildThisPage(K : IsamKeyStr; R : longint) : integer;
        var
            FreeRows       : integer;
            I              : integer;
            TRec           : RowRec;
        begin
            BuildThisPage := 2;
            with TRec do
            begin
                IKS := K;
                Ref := R;
                F_Key(IFBPtr, KeyNr, Ref, IKS, - 1);
                if IsamErrorClass > 1 then
                    exit;
                if (IsamErrorClass = 1) or (not KeyInBounds(IKS)) then
                begin
                    N_Key(IFBPtr, KeyNr, Ref, IKS);
                    if (IsamErrorClass = 1) or (not KeyInBounds(IKS)) then
                    begin
                        BuildThisPage := 1;
                        exit;
                    end;
                end;
            end;
            I := 1;
            while (I <= NrOfRows) and (KeyInBounds(TRec.IKS)) and (IsamErrorClass = 0) do
            begin
                BrowScreen[I] := TRec;
                with TRec do
                begin
                    inc(I);
                    if I <= NrOfRows then
                    begin
                        N_Key(IFBPtr, KeyNr, Ref, IKS);
                        if IsamErrorClass > 1 then
                            exit;
                    end;
                end;
            end;
            CurRow := 1;
            FreeRows := NrOfRows - LastRow;
            if FreeRows > 0 then
            begin
                if BuildPrevPage(FreeRows, Moved) > 1 then
                    exit;
                CurRow := Moved + 1;
            end;
            BuildThisPage := 0;
        end;

        function ActCurPage(Repos : boolean) : integer;
        var
            SaveRow        : integer;
        begin
            SaveRow := CurRow;
            EmptyBrowScreen;
            Result := BuildThisPage(KeyStr, Ref);
            if Result <> 0 then
            begin
                ActCurPage := Result;
                exit;
            end;
            if Repos then
                if SaveRow > CurRow then
                begin
                    Result := BuildPrevPage(SaveRow - CurRow, Moved);
                    if Result <> 0 then 
                    begin
                        ActCurPage := Result;
                        exit;
                    end;
                    inc(CurRow, Moved);
                end;
            if not CompleteaPage then
            begin
                ActCurPage := 2;
                exit;
            end;
            DisplayaPage;
            ActCurPage := 0;
        end;

        function FindLowHighRows(var LowRef, HighRef : longint) : integer;
            {-Return lowest and highest record numbers}
        var
            Result         : integer;
            IKS            : IsamKeyStr;
        begin
            IKS := LowKey;
            S_Key(IFBPtr, KeyNr, LowRef, IKS);
            if IsamErrorClass > 1 then 
            begin
                FindLowHighRows := 2;
                exit;
            end;
            if (IsamErrorClass = 1) or not KeyInBounds(IKS) then
            begin
                FindLowHighRows := 1;
                exit;
            end;
            IKS := HighKey;
            while length(IKS) < MaxKeyLen do
                IKS := IKS + #255;
            HighRef := $0FFFFFFF;
            F_Key(IFBPtr, KeyNr, HighRef, IKS, - 1);
            if IsamErrorClass > 1 then
            begin
                FindLowHighRows := 2;
                exit;
            end;
            if (IsamErrorClass = 1) or not KeyInBounds(IKS) then
            begin
                FindLowHighRows := 1;
                exit;
            end;
            FindLowHighRows := 0;
        end;

    begin                                         {BrowseI}
        if not KeyInBounds(KeyStr) then
            if KeyStr < LowKey then
                KeyStr := LowKey
            else
                KeyStr := HighKey;


        if NrOfRows > MaxRows then
            NrOfRows := MaxRows
        else if NrOfRows < MinRows then
            NrOfRows := MinRows;

        if NoNetMode then
        begin
            {Determine first and last acceptable records}
            Result := FindLowHighRows(LowRef, HighRef);
            if Result <> 0 then
            begin
                BrowseI := Result;
                exit;
            end;
        end;

        HorizOfs := 0;
        CurRow := 1;
        Result := ActCurPage(false);

        if Result <> 0 then
        begin
            BrowseI := Result;
            exit;
        end;

      {$IFDEF UseTpMouse}
      if MouseInstalled then
        SaveMouseOn := MouseCursorOn;
      if BrowseMouseEnabled then
      begin
        {Save mouse position and window}
        SaveMX := MouseWhereX+MouseXLo; {Absolute}
        SaveMY := MouseWhereY+MouseYLo;
        SaveMXL := MouseXLo+1;
        SaveMXH := MouseXHi;
        SaveMYL := MouseYLo+1;
        SaveMYH := MouseYHi;
        SaveWaitFor := WaitForButtonRelease;
        WaitForButtonRelease := True;
        HideMouse;
        {Set mouse window around the Browse list}
        with WindowP(W)^,Draw do
          if BrowseMouseScroll then
          begin
            {Let mouse move into frame}
            MouseWindow(XL, YL1, XH1, YH1);
            {Draw scroll marks}
            DrawMouseMarks(W);
            MouseDefaultXY(MouseXLo+1, MouseYLo+2, SaveMX, SaveMY);
          end else begin
            {Don't let mouse move into frame}
            MouseWindow(XL, YL, XH, YH);
            MouseDefaultXY(MouseXLo+1, MouseYLo+1, SaveMX, SaveMY);
          end;
        end;


        HideMouse;
        {$ENDIF}
        CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                        StartScreenRow, false, HorizOfs);
        BrowseI := 2;
        Action := ExitKey;
        repeat
            case Action of
                BKnone,
                BKchar : ;

                BKfirstRec :
                    {!!}
                    if not NoNetMode or (BrowScreen[1].Ref <> LowRef) then
                    begin
                        CurRow := 1;
                        EmptyBrowScreen;
                        Result := BuildFirstPage;
                        if Result <> 0 then
                        begin
                            BrowseI := Result;
                            exit;
                        end;
                        if not CompleteaPage then
                            exit;
                        DisplayaPage;
                    end
                    else
                    begin
                        CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                                        StartScreenRow, false, HorizOfs);
                        CurRow := 1;
                    end;

                BKlastRec :
                    {!!}
                    if not NoNetMode or (BrowScreen[LastRow].Ref <> HighRef) then
                    begin
                        CurRow := LastRow;
                        EmptyBrowScreen;
                        Result := BuildLastPage;
                        if Result <> 0 then
                        begin
                            BrowseI := Result;
                            exit;
                        end;
                        if not CompleteaPage then
                            exit;
                        DisplayaPage;
                    end
                    else
                    begin
                        CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                                        StartScreenRow, false, HorizOfs);
                        CurRow := LastRow;
                    end;

                BKpageUp :
                    {!!}
                    if not NoNetMode or (BrowScreen[1].Ref <> LowRef) then
                    begin
                        Result := BuildPrevPage(NrOfRows, Moved);
                        if Result <> 0 then
                        begin
                            BrowseI := Result;
                            exit;
                        end;
                        if not CompleteaPage then
                            exit;
                        DisplayaPage;
                    end
                    else
                    begin
                        CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                                        StartScreenRow, false, HorizOfs);
                        CurRow := 1;
                    end;

                BKpageDown : begin
                    {!!}
                    if not NoNetMode or (BrowScreen[LastRow].Ref <> HighRef) then
                    begin
                        Result := BuildNextPage(NrOfRows, Moved);
                        if Result <> 0 then
                        begin
                            BrowseI := Result;
                            exit;
                        end;
(*
            if Moved = 0 then
              CurRow := LastRow;
*)
                        if not CompleteaPage then
                            exit;
                        DisplayaPage;
                    end
                    else
                    begin
                        CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                                        StartScreenRow, false, HorizOfs);
                        CurRow := LastRow;
                    end;
                    end;
                BKup :
                    {!!}
                    if CurRow = 1 then
                    begin
                        if not NoNetMode or (BrowScreen[CurRow].Ref <> LowRef) then
                        begin
                            Result := BuildPrevPage(NrOfRows div 2, Moved);
                            if Result <> 0 then
                            begin
                                BrowseI := Result;
                                exit;
                            end;
                            if Moved <> 0 then
                                CurRow := Moved;
                            if not CompleteaPage then
                                exit;
                            DisplayaPage;
                        end;
                    end
                    else
                    begin
                        CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                                        StartScreenRow, false, HorizOfs);
                        dec(CurRow);
                    end;

                BKdown : begin

                    {!!}
                    if CurRow = LastRow then
                    begin
                        if not NoNetMode or (BrowScreen[CurRow].Ref <> HighRef) then
                        begin
                            Result := BuildNextPage(NrOfRows div 2, Moved);
                            if Result <> 0 then
                            begin
                                BrowseI := Result;
                                exit;
                            end;
                            if Moved <> 0 then
                                CurRow := LastRow + 1 - Moved;
                            if not CompleteaPage then
                                exit;
                            DisplayaPage;
                        end;
                    end
                    else
                    begin
                        CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                                        StartScreenRow, false, HorizOfs);
                        inc(CurRow);
                    end;
                    end;
                BKleft :
                    if HorizOfs > 0 then
                    begin
                        dec(HorizOfs);
                        DisplayaPage;
                    end;

                BKright :
                    begin
                        HorizOfs := (HorizOfs + 1) mod 32000;
                        DisplayaPage;
                    end;

                BKplus :
                    begin
                        Ref := BrowScreen[CurRow].Ref;
                        KeyStr := BrowScreen[CurRow].IKS;
                        Result := ActCurPage(true);
                        if Result <> 0 then
                        begin
                            BrowseI := Result;
                            exit;
                        end;
                    end;

                BKhelp :
                    if BrowseHelpPtr <> nil then
                        HelpRoutine(HelpForBrowse, IFBPtr, BrowseHelpIndex);

                BKProbe :
                    begin
                        CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                                        StartScreenRow, false, HorizOfs);
                        EvaluateMousePosition (W, Action, CurRow);
                    end;

                BKquit :
                    begin
                        ExitKey := BKquit;
                        BrowseI := 0;
                        exit;
                    end;
            end;                                     {Case}

            case Action of
                BKenter, BKuser0..BKuser9 :
                    begin
                        ExitKey := Action;
                        Ref := BrowScreen[CurRow].Ref;
                        KeyStr := BrowScreen[CurRow].IKS;
                        if Ref <> longint(- 1) then
                        begin
                            if VarRec then
                                GetVariableRec(IFBPtr, Ref, DatS, DatLen, Normal)
                            else if NetUsed then
                                GetNetRec(IFBPtr, Ref, DatS)
                            else
                                GetRec(IFBPtr, Ref, DatS);
                            if IsamErrorClass > 2 then
                                exit;
                            if (longint(DatS) <> longint(0)) or (IsamErrorClass = 2) then
                            begin
                                Ref := BrowScreen[CurRow].Ref;
                                KeyStr := BrowScreen[CurRow].IKS;
                                Result := ActCurPage(true);
                                if Result <> 0 then
                                begin
                                    BrowseI := Result;
                                    exit;
                                end;
                            end
                            else
                            begin
                                BrowseI := 0;
                                exit;
                            end;
                        end;
                    end;
            end;
{<==== new hilite turned on here }
            CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                            StartScreenRow, true, HorizOfs);

            case Action of
                BKtask0..BKtask9 :
                    if ProcSpecialTask <> nil then
                    begin
                        ExitKey := Action;
                        Ref := BrowScreen[CurRow].Ref;
                        if Ref <> longint(- 1) then
                        begin
                            if VarRec then
                                GetVariableRec(IFBPtr, Ref, DatS, DatLen, Normal)
                            else if NetUsed then
                                GetNetRec(IFBPtr, Ref, DatS)
                            else
                                GetRec(IFBPtr, Ref, DatS);
                            if IsamErrorClass > 2 then
                                exit;

                            if (longint(DatS) <> longint(0)) or (IsamErrorClass = 2) then
                            begin
                                Ref := BrowScreen[CurRow].Ref;
                                KeyStr := BrowScreen[CurRow].IKS;
                                CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                                                StartScreenRow, false, HorizOfs);
                                Result := ActCurPage(true);
                                if Result <> 0 then
                                begin
                                    BrowseI := Result;
                                    exit;
                                end;
                                CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                                                StartScreenRow, true, HorizOfs);
                                Action := BKnone;
                            end
                            else
                            begin
                                KeyStr := BrowScreen[CurRow].IKS;
                                CallSpecialTask(ProcSpecialTask, IFBPtr, DatS, Ref, KeyStr, KeyNr,
                                                Action, Result, DatLen);
                                if Result <> 0 then
                                begin
                                    BrowseI := Result;
                                    exit;
                                end;
                                Result := ActCurPage(true);
                                if Result <> 0 then
                                begin
                                    BrowseI := Result;
                                    exit;
                                end;
                            end;
                        end
                        else
                            Action := BKnone;
                    end;
            else
                begin
                    { this is the key/mouse command processor }
                    {$IFDEF UseTpMouse}
                    if BrowseMouseEnabled then
                        ShowMouse;
                    Action := GetCommand(BrowseKeySet, BrowseKeyPtr, ChWord);
                    if BrowseMouseEnabled then
                        HideMouse;
                    if Action = BKProbe then
                    begin
                       CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                                     StartScreenRow, false, HorizOfs);

                        EvaluateMousePosition(W, Action, CurRow);
                    end;
                    {$ELSE}
                    Action := GetCommand(BrowseKeySet, BrowseKeyPtr, ChWord);
                    {$ENDIF}

                end
            end;

      (* !! removed by KRK, 9/23/88
      CallDisplayaRow(ProcDisplayaRow, BrowScreen[CurRow], KeyNr, CurRow,
                  StartScreenRow, False, HorizOfs);
      *)
        until false;
    end;                                          {BrowseI}



    function BrowseM(IFBPtr         : IsamFileBlockPtr;
                     VarRec         : boolean;
                     KeyNr          : integer;
                     LowKey         : IsamKeyStr;
                     HighKey        : IsamKeyStr;

                     C1, R1, C2, R2 : integer;
                     Frame          : boolean;
                     Wattr, Fattr, Tattr : byte;
                     Title          : string;

                     var DatS;
                     var DatLen     : word;
                     var Ref        : longint;
                     var KeyStr     : IsamKeyStr;
                     var ExitKey    : BKtype;
                     ProcSpecialTask : pointer;
                     ProcBuildaRow  : pointer;
                     ProcDisplayaRow : pointer) : integer;

    var
        SaveTextAttr   : byte;
        SaveCursor     : word;
        SaveSFS        : boolean;
        SaveError      : integer;
        SaveMouse      : MouseStatePtr;


        function GetCursorMode : word;
        const
            Video          = $10;
            GetCurType     = $300;
        var
            Regs           : Registers;
        begin
            Regs.AX := GetCurType;
            Intr(Video, Regs);
            GetCursorMode := Regs.CX;
        end;

        procedure SetCursorMode(Mode : word);
        const
            Video          = $10;
            SetCurType     = $100;
        var
            Regs           : Registers;
        begin
            Regs.AX := SetCurType;
            Regs.CX := Mode;
            Intr(Video, Regs);
        end;



    begin                                        { Browse }

        SaveTextAttr := TextAttr;
        SaveCursor := GetCursorMode;

        {$IFDEF UseTpMouse}
        {Save the current mouse state, allocate storage for save-area}
        SaveMouseState(SaveMouse, true);
        {$ENDIF}
        {$IFDEF UseTpMouse}
        {Create the Window Needed for the Browse operation}
        if not MakeWindow(W, C1, R1, C2, R2, Frame, true, true,
                          Wattr, Fattr, Tattr, Title) then exit
        else
            if not DisplayWindow(W) then exit;
        {$ENDIF}

        {Hide the cursor}
        SetCursorMode($2000);

        TestSearchForSequential(IFBPtr, KeyNr, SaveSFS);
        EnableSearchForSequential(IFBPtr, KeyNr);
        BrowseM := BrowseI(IFBPtr, VarRec, KeyNr, LowKey, HighKey,
                           R1,
                           R2 - R1 + 1, DatS, DatLen, Ref, KeyStr, ExitKey,
                           ProcSpecialTask, ProcBuildaRow, ProcDisplayaRow,
                           (IFBPtr^.NSP <> nil));
        SaveError := IsamError;
        if SaveSFS then
            EnableSearchForSequential(IFBPtr, KeyNr)
        else
            DisableSearchForSequential(IFBPtr, KeyNr);
        IsamError := SaveError;
        IsamOK := (IsamError = 0);
        SetCursorMode(SaveCursor);
        TextAttr := SaveTextAttr;

        {$IFDEF UseTpMouse}
        {Erase and remove the browse window}
        DisposeWindow(EraseTopWIndow);
        {$ENDIF}

        {$IFDEF UseTpMouse}
        {Restore the current mouse state, de-allocate storage for save-area}
        RestoreMouseState(SaveMouse, true);
        {$ENDIF}

    end;                                         { Browse }



    function AddBrowseCommand(Cmd : BKtype; NumKeys : byte; Key1, Key2 : word) : boolean;
        {-Add a new command key assignment or change an existing one}
    begin
        AddBrowseCommand :=
        AddCommandPrim(BrowseKeySet, BrowseKeyMax, Cmd, NumKeys, Key1, Key2);
    end;

    procedure BuildaRow(var RR         : RowRec;
                        KeyNr          : integer;
                        var DatS;
                        DatLen         : word);

    var
        Structure :
            record
                Dele           : longint;
                S              : string[48];
            end absolute DatS;

    begin
        with RR do
        begin
            if Ref <> longint(- 1) then
            begin
                Row := Structure.S;
            end
            else
            begin
                Row := '';
                while length(Row) < MaxCols do
                begin
                    Row := Row + '**   ';
                end;
            end;
        end;
    end;


    procedure DisplayaRow(var RR         : RowRec;
                          KeyNr,
                          RowNr,
                          StartRow       : integer;
                          HighLight      : boolean;
                          var HorizOfs   : integer);

    const
        ScreenCols     = 80;
    var
        S              : string[255];
    begin
        if HorizOfs > 3 then
            HorizOfs := 3;
        if HighLight then
        begin
            TextColor(0);
            TextBackground(7);
        end
        else
        begin
            TextColor(15);
            TextBackground(0);
        end;
        S := RR.Row;
        S := copy(S, HorizOfs * 12 + 1, MaxCols);
        GotoXY(1, StartRow - 1 + RowNr);
        ClrEOL;
        write(copy(S, 1, ScreenCols));
    end;

    procedure SpecialTask(IFBPtr         : IsamFileBlockPtr;
                          var DatS;
                          Ref            : longint;
                          IKS            : IsamKeyStr;
                          KeyNr          : integer;
                          var Command    : BKtype;
                          var exitcode   : integer;
                          DatLen         : word);
    var
        Ch             : char;
    begin
        exitcode := 0;
        GotoXY(1, 10);
        ClrEOL;
        read(Ch);
        if Ch = 'P' then
            Command := BKpageDown
        else
            Command := BKnone;
    end;

begin
    { initialize procedure pointers }
    BrowseHelpPtr := nil;
    BrowseKeyPtr := @BrowseReadKey;
end.

